(* ==============================================================
// Copyright (c) 1996 by Aaron Castro and VK Thakur. 
// Version 3.0 enhancements by Eric Engler.
*)
unit Dbsearch;

interface
uses SysUtils, WinTypes, WinProcs, Messages, Classes, Graphics, Controls,
     Forms, DB, DBTables, DsgnIntf,
     Dbslocat, Sercheng, Pikflds, Dialogs;

type
   EDBSearchPatternNotFound = class(Exception);

type
   TdbSearch = class(TComponent)
   private
     { These items beginning with F are local versions
       of some of our properties }
     FLocateInfo: TLocateInfo;
     FDataSource: TDataSource;
     FDefSearchField: String;
     FHelpFile: String;
     FSearchFields: TStringList;
     function GetDataSource: TDataSource;
     procedure SetDataSource(ds: TDataSource);
     procedure SetSearchFields(FieldNames: TStringList);
     procedure InitLocateInfo;
     function SearchInMemo(Blob: TBlobField; Pattern: PChar; CaseSensitive,
                  ExactMatch: Boolean): Boolean;
     procedure SetMultiPattern(MultiPattern: string);
     function GetMultiPattern: string;
     procedure SetSinglePattern(SinglePattern: Char);
     function GetSinglePattern: Char;
   protected
     procedure DoLocate(LocateInfo: TLocateInfo);
   public
     { Our methods }
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function GetLocateInfo: boolean;
     procedure Locate;
     procedure LocateNext;
   published
     { These properties show in Object Inspector }
     property DataSource: TDataSource   read GetDataSource 
                                        write SetDataSource;
     property DefSearchField: string    read FDefSearchField 
                                        write FDefSearchField;
     property SearchFields: TStringList read FSearchFields
                                        write SetSearchFields;
     property MultiCharPattern: string  read GetMultiPattern 
                                        write SetMultiPattern;
     property SingleCharPattern: Char   read GetSinglePattern 
                                        write SetSinglePattern;
   end;

  TDefSrchFldProperty = class(TStringProperty)
      procedure GetValues(Proc: TGetStrProc); override;
      function GetAttributes: TPropertyAttributes; override;
  end;
  procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Samples', [TdbSearch]);
  RegisterComponentEditor(TdbSearch, TdbSearchEditor);
  RegisterPropertyEditor(TypeInfo(TStringList), TdbSearch, 'SearchFields', 
                         TSrchFldsProperty );
  RegisterPropertyEditor(TypeInfo(string), TdbSearch, 'DefSearchField', 
                         TDefSrchFldProperty );
end;

constructor TdbSearch.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FSearchFields := TStringList.Create;
   FlocateInfo.CaseSensitive:=FALSE;
   FlocateInfo.ExactMatch:=TRUE;
end;

destructor TdbSearch.Destroy;
begin
   inherited Destroy;
end;

procedure TdbSearch.SetMultiPattern(MultiPattern: string);
var
   Temp: PChar;
begin
   GetMem(Temp, Length(MultiPattern) + 3);
   StrPCopy(Temp, MultiPattern);
   SetMultiPat(Temp);
   FreeMem(Temp, Length(MultiPattern) + 3);
end;

function TdbSearch.GetMultiPattern: string;
begin
   Result := StrPas(GetMultiPat);
end;

procedure TdbSearch.SetSinglePattern(SinglePattern: Char);
begin
   SetSinglePat(SinglePattern);
end;

function TdbSearch.GetSinglePattern: Char;
begin
   Result := GetSinglePat;
end;

function TdbSearch.GetDataSource: TDataSource;
begin
   Result := FDataSource;
end;

procedure TdbSearch.SetDataSource(ds: TDatasource);
begin
  FDataSource := ds;
  InitLocateInfo;
end;

procedure TdbSearch.SetSearchFields(FieldNames: TStringList);
begin
  FSearchFields.Assign(FieldNames);
end;

procedure TdbSearch.InitLocateInfo;
begin
   with FLocateInfo do
   begin
      GotoFirst:=FALSE;
      GotoLast:=FALSE;
      FindNext:=FALSE;
   end;
end;

{ Called from "Locate" in this module just before calling DoLocate.
  If this returns FALSE, then user CANCELED the search }
function TdbSearch.GetLocateInfo: Boolean;
var
   loopctr: integer;
begin
   Result:=False;
   { Create the dialog box - see dbslocat.pas }
   DLGLocate := TDLGLocate.Create(Application);
   DLGLocate.rbPatternMatch.Caption := '&Allow Wildcards '
            + SingleCharPattern + ' and ' + MultiCharPattern;
   try
     if (FDataSource <> nil) then
     begin
       if FSearchFields.Count > 0 then
       begin
          for loopctr := 0 to FSearchFields.Count -1 do
             DLGLocate.AddLocateField(SearchFields[loopctr]);
       end
       else
         {Search is permitted on all table fields}
         for loopctr:=0 to FDataSource.DataSet.FieldCount - 1 do
            {Add only the ones the end-user can see}
            if (FDataSource.DataSet.Fields[loopctr].Visible) then
               DLGLocate.AddLocateField(
                    FDataSource.DataSet.Fields[loopctr].DisplayLabel);

       {--------------------------------------------}
       {   Initialize the dialogbox fields          }
       { Only use the default field if user didn't  }
       { already specify one.                       }
       If FLocateInfo.LocateField = '' then
          FLocateInfo.LocateField := FDefSearchField;

       { call func in dbslocat.pas                  }
       DLGLocate.SetLocateInfo(FLocateInfo);

       { ***** Show the main locate dialog ***** }
       if ( DLGLocate.ShowModal = mrOK ) then
       begin
          Result:=True;
          { Get the dialog results from func in dbslocat.pas }
          FLocateInfo:=DLGLocate.GetLocateInfo;
       end;
     end; { end "if (FDataSource <> nil)" }
   finally
     DLGLocate.Free;
   end; { end of "try...finally" }
end;

procedure TdbSearch.DoLocate(LocateInfo: TLocateInfo);

   { This function is local to DoLocate }
   function CanSearchOnIndex( FieldName: String ): Boolean;
   var
      I: Integer;
   begin
      Result := False;
      with (FDataSource.Dataset as TTable) do
      begin
         { refresh our index data }
         IndexDefs.Update;
         { see if the desired field is in any index }
         for I := 0 to IndexDefs.Count - 1 do
            if Pos(IndexDefs.Items[I].Fields, LocateInfo.LocateField) = 1 then
               Result := True;
      end;
   end; { end of CanSearchOnIndex }

{ Meat of DoLocate }
var
   Target, Pattern: Array[0..255] of Char;
   loopctr: Integer;
   oldpos: TBookmark;
   FieldNumber: Integer; {The index of the field that we are searching on}
   OldCursor: HCursor;
   StringMatch: Boolean;
   ActiveIndexFields: TStringList;
   ActiveIndex: String;
   I: Integer;
   label done_locate;

begin
   if LocateInfo.GotoFirst then
   begin
      with FDataSource.DataSet do
         First;
      goto done_locate;
   end;

   if LocateInfo.GotoLast then
   begin
      with FDataSource.DataSet do
         Last;
      goto done_locate;
   end;

   StrPCopy(Pattern, LocateInfo.LocateText);

   { Get the index of the field to search }
   for loopctr:=0 to FDataSource.Dataset.FieldCount - 1 do
      if FDataSource.Dataset.Fields[loopctr].DisplayLabel
       = LocateInfo.LocateField then
      begin
         FieldNumber:=loopctr;
         Break;
      end;

  if FieldNumber >= FDataSource.Dataset.FieldCount then
     raise EDBSearchPatternNotFound.Create('Search Field not in table');

  { Only use the index-based search if you're using TTable, and
    user has NOT specified wildcards }
  if (LocateInfo.ExactMatch or not IsPatternWild(Pattern))
   and (FDataSource.Dataset.ClassName = 'TTable')
   and not (FDatasource.Dataset as TTable).Database.IsSQLBased
   and CanSearchOnIndex(LocateInfo.LocateText) then
    With (FDataSource.Dataset as TTable) do
    begin
       { ==================================================== }
       { This is the driver routine for an index-based search }

       { Since this is a component, we assume the application
         program may be using a different index than the one
         we want to search with.  We'll save the current index,
         do the find with the index on desired column.
           Next, using the data from the target record, we'll
         populate the key fields of whatever index was being
         used.  We'll then get the same record using the orig
         index. }
       DisableControls;
       { Store the name of the currently active index }
       ActiveIndex := IndexName;
       ActiveIndexFields := TStringList.Create;
       try
          for I := 0 to IndexFieldCount - 1 do
             ActiveIndexFields.Add(IndexFields[I].FieldName);

          { Delphi will automatically pick an index based on the columns
            specified in IndexFieldNames - we know it will find one }
          IndexFieldNames := Fields[FieldNumber].FieldName;

          { This does the search - No need for a loop }
          if LocateInfo.FindNext then
             moveby(1)
          else
             { NOTE: FindKey always searches from the top of the file }
             FindKey([LocateInfo.LocateText]);

          { Replace field names with values }
          for I := 0 to ActiveIndexFields.Count - 1 do
             ActiveIndexFields[I]:= FieldByName(ActiveIndexFields[I]).AsString;

          { Get back to the original index and find the same record }
          IndexName := ActiveIndex;
          SetKey; { allows us to specify key values for current index }
          for I := 0 to ActiveIndexFields.Count - 1 do
             IndexFields[I].AsString := ActiveIndexFields[I];
          GoToKey; { directly access record }
       finally
          ActiveIndexFields.Free;
          EnableControls;
       end;  { end of "try...finally" block }
       Goto done_locate;
    end;    

    { ========================================================== }
    { This is the driver routine for the brute force search      }
    with FDataSource.DataSet do
    begin
       OldCursor:=SetCursor(LoadCursor(0, IDC_WAIT));
       DisableControls;
       { Save old position, in case we can't find the text }
       oldpos:=GetBookmark;
       { Start from first record in file, or the next one }
       if LocateInfo.FindNext = FALSE then
          First
       else
       begin
          Next;
          if eof then
          begin
             EnableControls;
             raise EDBSearchPatternNotFound.Create(
                'Search pattern: ' + LocateInfo.LocateText
                + ' not found in field: '
                + Fields[FieldNumber].FieldName);
          end;
       end;  { end else }

       { Loop to examine specified field of each record for a match }
       StringMatch := False;
       repeat
          if (Fields[FieldNumber].DataType = ftMemo) then
             { SearchInMemo is in this source file }
             StringMatch := SearchInMemo(TMemoField(Fields[FieldNumber]),
                Pattern, LocateInfo.CaseSensitive, not LocateInfo.ExactMatch)
          else
             { CompareStrings is in SerchEng.PAS }
             StringMatch := CompareStrings(StrPCopy(Target,
                Fields[FieldNumber].Text), Pattern, LocateInfo.CaseSensitive,
                not LocateInfo.ExactMatch);
          if StringMatch then
             Break
          else
             Next;
       until EOF;

       EnableControls;
       { End of the data and no match? reposition to previous position }
       if (not StringMatch) then
          Gotobookmark(oldpos);
       Freebookmark(oldpos);
    end;

    SetCursor(OldCursor);
    if (not StringMatch) then
       raise EDBSearchPatternNotFound.Create(
          'Search pattern: ' + LocateInfo.LocateText
          + ' not found in field: ' +
          FDataSource.DataSet.Fields[FieldNumber].FieldName);
done_locate:
end;

function TdbSearch.SearchInMemo(Blob: TBlobField; Pattern: PChar;
                     CaseSensitive, ExactMatch: Boolean): Boolean;
var
   BlobStream: TBlobStream;
   Buffer: Array[0..255] of Char;
   BytesRead: Byte;
   TotalRead: Word;
   P: PChar;

begin
(* What follows is convoluted, I agree. I was forced to resort to it since
   I just could not get the BlobStream.Read method to read data into
   a GetMemED PChar. Creating a large character array on the stack did not
   appear to be a good idea.  *)
   result := False;
   BlobStream := TBlobStream.Create(Blob, bmRead);
   { The largest single block that you can safely allocate on the
     heap is 65528 }
   GetMem(P, 65528);
   try
      BytesRead := 255;
      TotalRead := 0;
      repeat
         BytesRead := BlobStream.Read(Buffer, 255);
         Buffer[BytesRead] := Chr(0);
         StrCopy(P+TotalRead, Buffer);
         TotalRead := TotalRead + BytesRead;
      until (BytesRead < 255) or (TotalRead > 65273);
      (P + TotalRead)^ := Chr(0);
      Result := CompareStrings(P, Pattern, CaseSensitive, ExactMatch);
   finally
      BlobStream.Free;
      FreeMem(P, 65528);
   end; { end of "try...finally" }
end;

{ This method can be called from an application program. }
procedure TdbSearch.Locate;
begin
   InitLocateInfo;
   { Get info from dialog, then do search }
   if (GetLocateInfo) then
      DoLocate(FLocateInfo);
end;

{ This method can be called from an application program. }
{ The point in having this method is to skip the dialog
  box in case the application program itself has a
  "Find Next" button.  Normally, the application program
  won't have his own "Find Next" button - he'll just call the
  Locate method to show the dialog box again (it has a
  "Find Next" button). }
procedure TdbSearch.LocateNext;
begin
   FLocateInfo.FindNext:=FALSE; { default }
   { Do we know what to look for? }
   if (FLocateInfo.LocateText='') then
      Locate  { No, so bring up dialog }
   else
   begin
      { Locate from the current position }
      FLocateInfo.FindNext := TRUE;
      DoLocate(FLocateInfo);
   end; { end else }
end;

{ Return the default "Search Field" property attributes }
function TDefSrchFldProperty.GetAttributes: TPropertyAttributes;
begin
   Result := [paMultiSelect, paValueList, paSortList];
end;

procedure TDefSrchFldProperty.GetValues(Proc: TGetStrProc);
var
  I: Integer;
  TheComponent: TdbSearch;
begin
  TheComponent := GetComponent(0) as TdbSearch;
  if TheComponent.DataSource <> nil then
  begin
     if (TheComponent.SearchFields.Count > 0)then
         for I := 0 to TheComponent.SearchFields.Count - 1 do
            Proc(TheComponent.SearchFields[I])
     else
        for I:=0 to (TheComponent.DataSource.DataSet.FieldCount - 1) do
           Proc(TheComponent.DataSource.DataSet.Fields[I].DisplayLabel);
  end; { end if }
end;

end.
